"""Provide common pytest fixtures."""

import asyncio
from collections import deque
from collections.abc import Generator
from copy import deepcopy
import json
from typing import Any
from unittest.mock import AsyncMock, Mock, patch

from aiohttp import ClientSession, ClientWebSocketResponse
from aiohttp.http_websocket import WSMessage, WSMsgType
import pytest

from zwave_js_server.client import Client
from zwave_js_server.model.controller import Controller
from zwave_js_server.model.driver import Driver
from zwave_js_server.model.node import Node

from . import load_fixture
from .common import MockCommandProtocol

# pylint: disable=protected-access, unused-argument

TEST_URL = "ws://test.org:3000"


@pytest.fixture(name="controller_state", scope="session")
def controller_state_fixture() -> dict[str, Any]:
    """Load the controller state fixture data."""
    return json.loads(load_fixture("controller_state.json"))


@pytest.fixture(name="multisensor_6_state", scope="session")
def multisensor_6_state_fixture():
    """Load the multisensor 6 node state fixture data."""
    return json.loads(load_fixture("multisensor_6_state.json"))


@pytest.fixture(name="lock_schlage_be469_state", scope="session")
def lock_schlage_be469_state_fixture():
    """Load the schlage lock node state fixture data."""
    return json.loads(load_fixture("lock_schlage_be469_state.json"))


@pytest.fixture(name="timed_lock_state", scope="session")
def timed_lock_state_fixture() -> dict[str, Any]:
    """Load the timed lock node state fixture data."""
    return json.loads(load_fixture("timed_lock_state.json"))


@pytest.fixture(name="climate_radio_thermostat_ct100_plus_state", scope="session")
def climate_radio_thermostat_ct100_plus_state_fixture():
    """Load the radio thermostat node state fixture data."""
    return json.loads(load_fixture("climate_radio_thermostat_ct100_plus_state.json"))


@pytest.fixture(name="cover_qubino_shutter_state", scope="session")
def cover_qubino_shutter_state_fixture():
    """Load the qubino shutter cover node state fixture data."""
    return json.loads(load_fixture("cover_qubino_shutter_state.json"))


@pytest.fixture(name="idl_101_lock_state", scope="session")
def idl_101_lock_state_fixture():
    """Load the bad string meta data node state fixture data."""
    return json.loads(load_fixture("idl_101_lock_state.json"))


@pytest.fixture(name="wallmote_central_scene_state", scope="session")
def wallmote_central_scene_state_fixture():
    """Load the wallmote central scene node state fixture data."""
    return json.loads(load_fixture("wallmote_central_scene_state.json"))


@pytest.fixture(name="unparseable_json_string_value_state", scope="session")
def unparseable_json_string_value_state_fixture():
    """Load the unparseable string json value node state fixture data."""
    return json.loads(load_fixture("unparseable_json_string_value_state.json"))


@pytest.fixture(name="partial_and_full_parameter_state", scope="session")
def partial_and_full_parameter_state_fixture():
    """Load the node that has both partial and full parameters state fixture data."""
    return json.loads(load_fixture("partial_and_full_parameter_state.json"))


@pytest.fixture(name="invalid_multilevel_sensor_type_state", scope="session")
def invalid_multilevel_sensor_type_state_fixture():
    """Load the node that has an invalid multilevel sensor type state fixture data."""
    return json.loads(load_fixture("invalid_multilevel_sensor_type_state.json"))


@pytest.fixture(name="inovelli_switch_state", scope="session")
def inovelli_switch_state_fixture():
    """Load the bad string meta data node state fixture data."""
    return json.loads(load_fixture("inovelli_switch_state.json"))


@pytest.fixture(name="ring_keypad_state", scope="session")
def ring_keypad_state_fixture():
    """Load the ring keypad node state fixture data."""
    return json.loads(load_fixture("ring_keypad_state.json"))


@pytest.fixture(name="endpoints_with_command_classes_state", scope="session")
def endpoints_with_command_classes_state_fixture():
    """Load the node state fixture data with command classes on the endpoint."""
    return json.loads(load_fixture("endpoints_with_command_classes_state.json"))


@pytest.fixture(name="switch_enbrighten_zw3010_state", scope="session")
def switch_enbrighten_zw3010_state_fixture():
    """Load the enbrighten zw3010 switch node state fixture data."""
    return json.loads(load_fixture("switch_enbrighten_zw3010_state.json"))


@pytest.fixture(name="energy_production_state", scope="session")
def energy_production_state_fixture():
    """Load a mock node with energy production CC state fixture data."""
    return json.loads(load_fixture("energy_production_state.json"))


@pytest.fixture(name="lock_ultraloq_ubolt_pro_state", scope="session")
def lock_ultraloq_ubolt_pro_state_fixture():
    """Load the ultraloq U-Bolt Pro lock state fixture data."""
    return json.loads(load_fixture("lock_ultraloq_ubolt_pro_state.json"))


@pytest.fixture(name="device_config", scope="session")
def device_config_fixture() -> dict[str, Any]:
    """Load the device config fixture data."""
    return json.loads(load_fixture("device_config.json"))


@pytest.fixture(name="client_session")
def client_session_fixture(ws_client: AsyncMock) -> AsyncMock:
    """Mock an aiohttp client session."""
    client_session = AsyncMock(spec_set=ClientSession)
    client_session.ws_connect.side_effect = AsyncMock(return_value=ws_client)
    return client_session


def create_ws_message(result: dict[str, Any]) -> Mock:
    """Return a mock WSMessage."""
    message = Mock(spec_set=WSMessage)
    message.type = WSMsgType.TEXT
    message.data = json.dumps(result)
    message.json.return_value = result
    return message


@pytest.fixture(name="messages")
def messages_fixture() -> deque[Mock]:
    """Return a message buffer for the WS client."""
    return deque()


@pytest.fixture(name="ws_client")
async def ws_client_fixture(
    version_data: dict[str, Any],
    ws_message: Mock,
    result: dict[str, Any],
    messages: deque[Mock],
    initialize_data: dict[str, Any],
    get_log_config_data: dict[str, Any],
) -> AsyncMock:
    """Mock a websocket client.

    This fixture only allows a single message to be received.
    """
    ws_client = AsyncMock(spec_set=ClientWebSocketResponse, closed=False)
    ws_client.receive_json.side_effect = (
        version_data,
        initialize_data,
        get_log_config_data,
        result,
    )
    for data in (version_data, initialize_data, get_log_config_data, result):
        messages.append(create_ws_message(data))

    async def receive() -> Mock:
        """Return a websocket message."""
        await asyncio.sleep(0)

        try:
            message = messages.popleft()
        except IndexError:
            ws_client.closed = True
            return WSMessage(WSMsgType.CLOSED, None, None)

        return message

    ws_client.receive.side_effect = receive

    async def close_client(msg: dict[str, Any]) -> None:
        """Close the client."""
        if msg["command"] in ("initialize", "start_listening"):
            return

        # We only want to skip for the initial call
        if (
            msg["command"] == "driver.get_log_config"
            and msg["messageId"] == "get-initial-log-config"
        ):
            return

        await asyncio.sleep(0)
        ws_client.closed = True

    ws_client.send_json.side_effect = close_client

    async def reset_close() -> None:
        """Reset the websocket client close method."""
        ws_client.closed = True

    ws_client.close.side_effect = reset_close

    return ws_client


@pytest.fixture(name="await_other")
async def await_other_fixture():
    """Await all other task but the current task."""

    async def wait_for_tasks(current_task):
        """Wait for the tasks."""
        tasks = asyncio.all_tasks() - {current_task}
        await asyncio.gather(*tasks)

    return wait_for_tasks


@pytest.fixture(name="driver_ready")
async def driver_ready_fixture():
    """Return an asyncio.Event for driver ready."""
    return asyncio.Event()


@pytest.fixture(name="version_data")
def version_data_fixture() -> dict[str, Any]:
    """Return mock version data."""
    return {
        "type": "version",
        "driverVersion": "test_driver_version",
        "serverVersion": "test_server_version",
        "homeId": "test_home_id",
        "minSchemaVersion": 0,
        "maxSchemaVersion": 44,
    }


@pytest.fixture(name="initialize_data")
def initialize_data_fixture() -> dict[str, Any]:
    """Return mock initialize data."""
    return {
        "type": "result",
        "success": True,
        "result": {},
        "messageId": "initialize",
    }


@pytest.fixture(name="log_config")
def log_config_fixture() -> dict[str, Any]:
    """Return log config."""
    return {
        "enabled": True,
        "level": "info",
        "logToFile": False,
        "filename": "",
        "forceConsole": False,
    }


@pytest.fixture(name="get_log_config_data")
def get_log_config_data_fixture(log_config: dict[str, Any]) -> dict[str, Any]:
    """Return mock get_log_config data."""
    return {
        "type": "result",
        "success": True,
        "result": {"config": log_config},
        "messageId": "get-initial-log-config",
    }


@pytest.fixture(name="url")
def url_fixture():
    """Return a test url."""
    return TEST_URL


@pytest.fixture(name="result")
def result_fixture(controller_state: dict[str, Any], uuid4: str) -> dict[str, Any]:
    """Return a server result message."""
    return {
        "type": "result",
        "success": True,
        "result": {"state": controller_state},
        "messageId": uuid4,
    }


@pytest.fixture(name="ws_message")
def ws_message_fixture(result: dict[str, Any]) -> Mock:
    """Return a mock WSMessage."""
    return create_ws_message(result)


@pytest.fixture(name="uuid4")
def mock_uuid_fixture() -> Generator[str, None, None]:
    """Patch uuid4."""
    uuid4_hex = "1234"
    with patch("uuid.uuid4") as uuid4:
        uuid4.return_value.hex = uuid4_hex
        yield uuid4_hex


@pytest.fixture(name="client")
async def client_fixture(
    client_session: AsyncMock,
    ws_client: AsyncMock,
    uuid4: str,
) -> Client:
    """Return a client with a mock websocket transport.

    This fixture needs to be a coroutine function to get an event loop
    when creating the client.
    """
    client = Client("ws://test.org", client_session)
    client._client = ws_client
    return client


@pytest.fixture(name="mock_command")
def mock_command_fixture(
    ws_client: AsyncMock, client: Client, uuid4: str
) -> MockCommandProtocol:
    """Mock a command and response."""
    mock_responses: list[tuple[dict, dict, bool]] = []
    ack_commands: list[dict] = []

    def apply_mock_command(
        match_command: dict, response: dict, success: bool = True
    ) -> list[dict]:
        """Apply the mock command and response return value to the transport.

        Return the list with correctly acknowledged commands.
        """
        mock_responses.append((match_command, response, success))
        return ack_commands

    async def set_response(message: dict[str, Any]) -> None:
        """Check the message and set the mocked response if a command matches."""
        for match_command, response, success in mock_responses:
            if all(message[key] == value for key, value in match_command.items()):
                ack_commands.append(message)
                received_message = {
                    "type": "result",
                    "messageId": uuid4,
                    "success": success,
                }
                if success:
                    received_message["result"] = response
                else:
                    received_message.update(response)
                client._handle_incoming_message(received_message)
                return

        raise RuntimeError("Command not mocked!")

    ws_client.send_json.side_effect = set_response

    return apply_mock_command


@pytest.fixture(name="driver")
def driver_fixture(
    client: Client,
    controller_state: dict[str, Any],
    log_config: dict[str, Any],
) -> Driver:
    """Return a driver instance with a supporting client."""
    client.driver = Driver(client, deepcopy(controller_state), log_config)
    return client.driver


@pytest.fixture(name="multisensor_6")
def multisensor_6_fixture(driver, multisensor_6_state):
    """Mock a multisensor 6 node."""
    node = Node(driver.client, deepcopy(multisensor_6_state))
    driver.controller.nodes[node.node_id] = node
    return node


@pytest.fixture(name="lock_schlage_be469")
def lock_schlage_be469_fixture(driver, lock_schlage_be469_state):
    """Mock a schlage lock node."""
    node = Node(driver.client, deepcopy(lock_schlage_be469_state))
    driver.controller.nodes[node.node_id] = node
    return node


@pytest.fixture(name="timed_lock")
def timed_lock_fixture(driver: Driver, timed_lock_state: dict[str, Any]) -> Node:
    """Mock a schlage lock node."""
    node = Node(driver.client, deepcopy(timed_lock_state))
    driver.controller.nodes[node.node_id] = node
    return node


@pytest.fixture(name="climate_radio_thermostat_ct100_plus")
def climate_radio_thermostat_ct100_plus_fixture(
    driver, climate_radio_thermostat_ct100_plus_state
):
    """Mock a radio thermostat node."""
    node = Node(driver.client, deepcopy(climate_radio_thermostat_ct100_plus_state))
    driver.controller.nodes[node.node_id] = node
    return node


@pytest.fixture(name="cover_qubino_shutter")
def cover_qubino_shutter_fixture(driver, cover_qubino_shutter_state):
    """Mock a qubino shutter cover node."""
    node = Node(driver.client, deepcopy(cover_qubino_shutter_state))
    driver.controller.nodes[node.node_id] = node
    return node


@pytest.fixture(name="wallmote_central_scene")
def wallmote_central_scene_fixture(driver, wallmote_central_scene_state):
    """Mock a wallmote central scene node."""
    node = Node(driver.client, deepcopy(wallmote_central_scene_state))
    driver.controller.nodes[node.node_id] = node
    return node


@pytest.fixture(name="controller")
def controller_fixture(driver, controller_state):
    """Return a controller instance with a supporting client."""
    controller = Controller(driver.client, deepcopy(controller_state))
    return controller


@pytest.fixture(name="inovelli_switch")
def inovelli_switch_fixture(driver, inovelli_switch_state):
    """Mock a inovelli switch node."""
    node = Node(driver.client, deepcopy(inovelli_switch_state))
    driver.controller.nodes[node.node_id] = node
    return node


@pytest.fixture(name="ring_keypad")
def ring_keypad_fixture(driver, ring_keypad_state):
    """Mock a ring keypad node."""
    node = Node(driver.client, deepcopy(ring_keypad_state))
    driver.controller.nodes[node.node_id] = node
    return node


@pytest.fixture(name="partial_and_full_parameter")
def partial_and_full_parameter_fixture(driver, partial_and_full_parameter_state):
    """Mock a node that has both partial and full parameters."""
    node = Node(driver.client, deepcopy(partial_and_full_parameter_state))
    driver.controller.nodes[node.node_id] = node
    return node


@pytest.fixture(name="invalid_multilevel_sensor_type")
def invalid_multilevel_sensor_type_fixture(
    driver, invalid_multilevel_sensor_type_state
):
    """Mock a node that has invalid multilevel sensor type."""
    node = Node(driver.client, deepcopy(invalid_multilevel_sensor_type_state))
    driver.controller.nodes[node.node_id] = node
    return node


@pytest.fixture(name="endpoints_with_command_classes")
def endpoints_with_command_classes_fixture(
    driver, endpoints_with_command_classes_state
):
    """Mock a node with command classes on an endpoint."""
    node = Node(driver.client, deepcopy(endpoints_with_command_classes_state))
    driver.controller.nodes[node.node_id] = node
    return node


@pytest.fixture(name="switch_enbrighten_zw3010")
def switch_enbrighten_zw3010_fixture(driver, switch_enbrighten_zw3010_state):
    """Mock an Enbrighten ZW3010 switch node."""
    node = Node(driver.client, deepcopy(switch_enbrighten_zw3010_state))
    driver.controller.nodes[node.node_id] = node
    return node


@pytest.fixture(name="energy_production")
def energy_prodution_fixture(driver, energy_production_state):
    """Mock a mock node with Energy Production CC."""
    node = Node(driver.client, deepcopy(energy_production_state))
    driver.controller.nodes[node.node_id] = node
    return node


@pytest.fixture(name="lock_ultraloq_ubolt_pro")
def lock_ultraloq_ubolt_pro_fixture(driver, lock_ultraloq_ubolt_pro_state):
    """Mock an Ultraloq U-Bolt Pro lock node."""
    node = Node(driver.client, deepcopy(lock_ultraloq_ubolt_pro_state))
    driver.controller.nodes[node.node_id] = node
    return node



"""
Model for a Zwave Node's device class.

https://zwave-js.github.io/node-zwave-js/#/api/node?id=deviceclass
"""

from __future__ import annotations

from dataclasses import dataclass
from typing import TypedDict


class DeviceClassItemDataType(TypedDict):
    """Represent a device class data dict type."""

    key: int
    label: str


class DeviceClassDataType(TypedDict):
    """Represent a device class data dict type."""

    basic: DeviceClassItemDataType
    generic: DeviceClassItemDataType
    specific: DeviceClassItemDataType


@dataclass
class DeviceClassItem:
    """Model for a DeviceClass item (e.g. basic or generic)."""

    key: int
    label: str


class DeviceClass:
    """Model for a Zwave Node's device class."""

    def __init__(self, data: DeviceClassDataType) -> None:
        """Initialize."""
        self._basic = DeviceClassItem(
            key=data["basic"]["key"],
            label=data["basic"]["label"],
        )
        self._generic = DeviceClassItem(
            key=data["generic"]["key"],
            label=data["generic"]["label"],
        )
        self._specific = DeviceClassItem(
            key=data["specific"]["key"],
            label=data["specific"]["label"],
        )

    @property
    def basic(self) -> DeviceClassItem:
        """Return basic DeviceClass."""
        return self._basic

    @property
    def generic(self) -> DeviceClassItem:
        """Return generic DeviceClass."""
        return self._generic

    @property
    def specific(self) -> DeviceClassItem:
        """Return specific DeviceClass."""
        return self._specific



"""Firmware update helper."""

from __future__ import annotations

import asyncio

import aiohttp

from .client import Client
from .model.driver.firmware import DriverFirmwareUpdateData, DriverFirmwareUpdateResult
from .model.node import Node
from .model.node.firmware import NodeFirmwareUpdateData, NodeFirmwareUpdateResult


async def update_firmware(
    url: str,
    node: Node,
    updates: list[NodeFirmwareUpdateData],
    session: aiohttp.ClientSession,
    additional_user_agent_components: dict[str, str] | None = None,
) -> NodeFirmwareUpdateResult:
    """Send updateFirmware command to Node."""
    client = Client(
        url, session, additional_user_agent_components=additional_user_agent_components
    )
    await client.connect()
    await client.initialize()

    receive_task = asyncio.get_running_loop().create_task(client.receive_until_closed())

    cmd = {
        "command": "node.update_firmware",
        "nodeId": node.node_id,
        "updates": [update.to_dict() for update in updates],
    }

    data = await client.async_send_command(cmd, require_schema=29)
    await client.disconnect()
    if not receive_task.done():
        receive_task.cancel()

    return NodeFirmwareUpdateResult(node, data["result"])


async def driver_firmware_update_otw(
    url: str,
    firmware_file: DriverFirmwareUpdateData,
    session: aiohttp.ClientSession,
    additional_user_agent_components: dict[str, str] | None = None,
) -> DriverFirmwareUpdateResult:
    """
    Send firmwareUpdateOTW command to Driver.

    Sending the wrong firmware to a driver can brick it and make it unrecoverable.
    Consumers of this library should build mechanisms to ensure that users understand
    the risks.
    """
    client = Client(
        url, session, additional_user_agent_components=additional_user_agent_components
    )
    await client.connect()
    await client.initialize()

    receive_task = asyncio.get_running_loop().create_task(client.receive_until_closed())

    data = await client.async_send_command(
        {
            "command": "driver.firmware_update_otw",
            **firmware_file.to_dict(),
        },
        require_schema=29,
    )
    await client.disconnect()
    if not receive_task.done():
        receive_task.cancel()

    return DriverFirmwareUpdateResult(data["result"])



"""Test lock utility functions."""

import copy

import pytest

from test.common import MockCommandProtocol
from zwave_js_server.const import SupervisionStatus
from zwave_js_server.const.command_class.lock import (
    ATTR_CODE_SLOT,
    ATTR_IN_USE,
    ATTR_NAME,
    ATTR_USERCODE,
    LOCK_USERCODE_ID_PROPERTY,
    LOCK_USERCODE_PROPERTY,
    LOCK_USERCODE_STATUS_PROPERTY,
    CodeSlotStatus,
    DoorLockCCConfigurationSetOptions,
    OperationType,
)
from zwave_js_server.exceptions import NotFoundError
from zwave_js_server.model.node import Node
from zwave_js_server.model.value import SupervisionResult
from zwave_js_server.util.lock import (
    clear_usercode,
    get_code_slots,
    get_usercode,
    get_usercode_from_node,
    get_usercodes,
    set_configuration,
    set_usercode,
    set_usercodes,
)

from .const import CODE_SLOTS


def test_get_code_slots(lock_schlage_be469):
    """Test get_code_slots utility function."""
    node = lock_schlage_be469
    assert get_code_slots(node) == [
        {k: v for k, v in code_slot.items() if k != ATTR_USERCODE}
        for code_slot in CODE_SLOTS
    ]


def test_get_usercode(lock_schlage_be469):
    """Test get_usercode utility function."""
    node = lock_schlage_be469

    # Test in use slot
    slot = get_usercode(node, 1)
    assert all(char == "*" for char in slot[ATTR_USERCODE])

    # Test unused slot
    assert get_usercode(node, 29)[ATTR_USERCODE] == ""

    # Test unknown slot
    assert get_usercode(node, 30)[ATTR_USERCODE] is None

    # Test invalid slot
    with pytest.raises(NotFoundError):
        get_usercode(node, 100)


def test_get_usercodes(lock_schlage_be469):
    """Test get_usercodes utility function."""
    node = lock_schlage_be469
    assert get_usercodes(node) == CODE_SLOTS


async def test_set_usercode(lock_schlage_be469, mock_command, uuid4):
    """Test set_usercode utility function."""
    node = lock_schlage_be469
    ack_commands = mock_command(
        {"command": "node.set_value", "nodeId": node.node_id},
        {"result": {"status": 255}},
    )

    # Test valid code
    await set_usercode(node, 1, "1234")
    assert len(ack_commands) == 1
    assert ack_commands[0] == {
        "command": "node.set_value",
        "nodeId": 20,
        "messageId": uuid4,
        "valueId": {
            "commandClass": 99,
            "endpoint": 0,
            "property": "userCode",
            "propertyKey": 1,
        },
        "value": "1234",
    }

    # Test invalid code slot
    with pytest.raises(NotFoundError):
        await set_usercode(node, 100, "1234")

    # assert no new command calls
    assert len(ack_commands) == 1


async def test_set_usercodes(
    lock_schlage_be469: Node, mock_command: MockCommandProtocol, uuid4: str
) -> None:
    """Test set_usercodes utility function."""
    node = lock_schlage_be469
    ack_commands = mock_command(
        {"command": "endpoint.invoke_cc_api", "endpoint": 0, "nodeId": node.node_id},
        {"response": {"status": 255}},
    )

    assert await set_usercodes(node, {1: "1234"}) == SupervisionResult(
        {"status": SupervisionStatus.SUCCESS}
    )
    assert len(ack_commands) == 1
    assert ack_commands[0] == {
        "command": "endpoint.invoke_cc_api",
        "commandClass": 99,
        "endpoint": 0,
        "methodName": "setMany",
        "nodeId": 20,
        "messageId": uuid4,
        "args": [
            [
                {
                    LOCK_USERCODE_STATUS_PROPERTY: CodeSlotStatus.ENABLED,
                    LOCK_USERCODE_ID_PROPERTY: 1,
                    LOCK_USERCODE_PROPERTY: "1234",
                }
            ]
        ],
    }


async def test_set_usercodes_invalid(
    lock_schlage_be469: Node, mock_command: MockCommandProtocol
) -> None:
    """Test set_usercodes utility function with invalid response."""
    node = lock_schlage_be469
    mock_command(
        {"command": "endpoint.invoke_cc_api", "endpoint": 0, "nodeId": node.node_id},
        {"response": {}},
    )

    with pytest.raises(ValueError):
        assert await set_usercodes(node, {"1": 1234})


async def test_clear_usercode(lock_schlage_be469, mock_command, uuid4):
    """Test clear_usercode utility function."""
    node = lock_schlage_be469
    ack_commands = mock_command(
        {"command": "node.set_value", "nodeId": node.node_id},
        {"result": {"status": 255}},
    )

    # Test valid code
    await clear_usercode(node, 1)
    assert len(ack_commands) == 1
    assert ack_commands[0] == {
        "command": "node.set_value",
        "nodeId": 20,
        "messageId": uuid4,
        "valueId": {
            "commandClass": 99,
            "endpoint": 0,
            "property": "userIdStatus",
            "propertyKey": 1,
        },
        "value": 0,
    }

    # Test invalid code slot
    with pytest.raises(NotFoundError):
        await clear_usercode(node, 100)

    # assert no new command calls
    assert len(ack_commands) == 1


async def test_get_usercode_from_node(lock_schlage_be469, mock_command, uuid4):
    """Test get_usercode_from_node utility function."""
    node = lock_schlage_be469
    ack_commands = mock_command(
        {"command": "endpoint.invoke_cc_api", "nodeId": node.node_id, "endpoint": 0},
        {"response": {"userIdStatus": 1, "userCode": "**********"}},
    )

    # Test valid code
    assert await get_usercode_from_node(node, 1) == {
        ATTR_NAME: "User Code (1)",
        ATTR_CODE_SLOT: 1,
        ATTR_IN_USE: True,
        ATTR_USERCODE: "**********",
    }
    assert len(ack_commands) == 1
    assert ack_commands[0] == {
        "command": "endpoint.invoke_cc_api",
        "nodeId": 20,
        "endpoint": 0,
        "commandClass": 99,
        "messageId": uuid4,
        "methodName": "get",
        "args": [1],
    }


async def test_set_configuration_empty_response(
    lock_schlage_be469, mock_command, uuid4
):
    """Test set_configuration utility function without response."""
    node = lock_schlage_be469
    ack_commands = mock_command(
        {"command": "endpoint.invoke_cc_api", "nodeId": node.node_id, "endpoint": 0},
        {},
    )
    await set_configuration(
        node.endpoints[0],
        DoorLockCCConfigurationSetOptions(OperationType.CONSTANT),
    )

    assert len(ack_commands) == 1
    assert ack_commands[0] == {
        "command": "endpoint.invoke_cc_api",
        "nodeId": 20,
        "endpoint": 0,
        "commandClass": 98,
        "messageId": uuid4,
        "methodName": "setConfiguration",
        "args": [
            {
                "insideHandlesCanOpenDoorConfiguration": [True, True, True, True],
                "operationType": 1,
                "outsideHandlesCanOpenDoorConfiguration": [True, True, True, True],
            }
        ],
    }

    with pytest.raises(ValueError):
        await set_configuration(
            node.endpoints[0],
            DoorLockCCConfigurationSetOptions(OperationType.CONSTANT, 1),
        )

    with pytest.raises(ValueError):
        await set_configuration(
            node.endpoints[0],
            DoorLockCCConfigurationSetOptions(OperationType.TIMED),
        )


async def test_set_configuration_with_response(lock_schlage_be469, mock_command, uuid4):
    """Test set_configuration utility function with response."""
    node = lock_schlage_be469
    ack_commands = mock_command(
        {"command": "endpoint.invoke_cc_api", "nodeId": node.node_id, "endpoint": 0},
        {"response": {"status": 0}},
    )
    result = await set_configuration(
        node.endpoints[0],
        DoorLockCCConfigurationSetOptions(OperationType.CONSTANT),
    )
    assert result.status == SupervisionStatus.NO_SUPPORT

    assert len(ack_commands) == 1
    assert ack_commands[0] == {
        "command": "endpoint.invoke_cc_api",
        "nodeId": 20,
        "endpoint": 0,
        "commandClass": 98,
        "messageId": uuid4,
        "methodName": "setConfiguration",
        "args": [
            {
                "insideHandlesCanOpenDoorConfiguration": [True, True, True, True],
                "operationType": 1,
                "outsideHandlesCanOpenDoorConfiguration": [True, True, True, True],
            }
        ],
    }


async def test_set_configuration_v4(
    driver, lock_ultraloq_ubolt_pro_state, mock_command, uuid4
):
    """Test v4 properties in set_configuration utility function."""
    node_state = copy.deepcopy(lock_ultraloq_ubolt_pro_state)
    node_state["values"].remove(
        {
            "endpoint": 0,
            "commandClass": 98,
            "commandClassName": "Door Lock",
            "property": "twistAssist",
            "propertyName": "twistAssist",
            "ccVersion": 4,
            "metadata": {
                "type": "boolean",
                "readable": True,
                "writeable": True,
                "label": "Twist Assist enabled",
            },
            "value": False,
        },
    )
    node = Node(driver.client, node_state)
    ack_commands = mock_command(
        {"command": "endpoint.invoke_cc_api", "nodeId": node.node_id, "endpoint": 0},
        {},
    )
    await set_configuration(
        node.endpoints[0],
        DoorLockCCConfigurationSetOptions(
            OperationType.CONSTANT, hold_and_release_time=2, block_to_block=True
        ),
    )

    assert len(ack_commands) == 1
    assert ack_commands[0] == {
        "command": "endpoint.invoke_cc_api",
        "nodeId": 34,
        "endpoint": 0,
        "commandClass": 98,
        "messageId": uuid4,
        "methodName": "setConfiguration",
        "args": [
            {
                "insideHandlesCanOpenDoorConfiguration": [True, True, True, True],
                "operationType": 1,
                "outsideHandlesCanOpenDoorConfiguration": [True, True, True, True],
                "autoRelockTime": 0,
                "holdAndReleaseTime": 2,
                "blockToBlock": True,
            }
        ],
    }

    # Test a property that isn't on the node
    with pytest.raises(ValueError):
        await set_configuration(
            node.endpoints[0],
            DoorLockCCConfigurationSetOptions(
                OperationType.CONSTANT, twist_assist=True
            ),
        )


@pytest.mark.usefixtures("driver")
async def test_set_configuration_timed_lock(
    timed_lock: Node, mock_command: MockCommandProtocol, uuid4: str
) -> None:
    """Test a timed lock in set_configuration utility function."""
    node = timed_lock
    ack_commands = mock_command(
        {"command": "endpoint.invoke_cc_api", "nodeId": node.node_id, "endpoint": 0},
        {},
    )
    await set_configuration(
        node.endpoints[0],
        DoorLockCCConfigurationSetOptions(
            OperationType.TIMED, lock_timeout_configuration=42
        ),
    )

    assert len(ack_commands) == 1
    assert ack_commands[0] == {
        "command": "endpoint.invoke_cc_api",
        "nodeId": 7,
        "endpoint": 0,
        "commandClass": 98,
        "messageId": uuid4,
        "methodName": "setConfiguration",
        "args": [
            {
                "operationType": OperationType.TIMED,
                "autoRelockTime": 0,
                "blockToBlock": True,
                "holdAndReleaseTime": 0,
                "twistAssist": True,
                "insideHandlesCanOpenDoorConfiguration": [True, True, True, True],
                "outsideHandlesCanOpenDoorConfiguration": [True, True, True, True],
                "lockTimeoutConfiguration": 42,
            }
        ],
    }



"""Generic Utility helper functions."""

from __future__ import annotations

import base64
import json
from typing import Any

from ..exceptions import UnparseableValue


def is_json_string(value: Any) -> bool:
    """Check if the provided string looks like json."""
    # NOTE: we do not use json.loads here as it is not strict enough
    return isinstance(value, str) and value.startswith("{") and value.endswith("}")


def convert_bytes_to_base64(data: bytes) -> str:
    """Convert bytes data to base64 for serialization."""
    return base64.b64encode(data).decode("ascii")


def convert_base64_to_bytes(data: str) -> bytes:
    """Convert base64 data to bytes for deserialization."""
    return base64.b64decode(data)


def parse_buffer(value: dict[str, Any] | str) -> str:
    """Parse value from a buffer data type."""
    if isinstance(value, dict):
        return parse_buffer_from_dict(value)

    if is_json_string(value):
        return parse_buffer_from_json(value)

    return value


def parse_buffer_from_dict(value: dict[str, Any]) -> str:
    """Parse value dictionary from a buffer data type."""
    if value.get("type") != "Buffer" or "data" not in value:
        raise UnparseableValue(f"Unparseable value: {value}") from ValueError(
            "JSON does not match expected schema"
        )
    return "".join([chr(x) for x in value["data"]])


def parse_buffer_from_json(value: str) -> str:
    """Parse value string from a buffer data type."""
    try:
        return parse_buffer_from_dict(json.loads(value))
    except ValueError as err:
        raise UnparseableValue(f"Unparseable value: {value}") from err



"""Common models for statistics."""

from __future__ import annotations

from dataclasses import dataclass, field
from functools import cached_property
from typing import TYPE_CHECKING, TypedDict

from zwave_js_server.exceptions import RepeaterRssiErrorReceived, RssiErrorReceived

from ..const import ProtocolDataRate, RssiError

if TYPE_CHECKING:
    from ..client import Client
    from .node import Node


class RouteStatisticsDataType(TypedDict, total=False):
    """Represent a route statistics data dict type."""

    protocolDataRate: int
    repeaters: list[int]
    rssi: int
    repeaterRSSI: list[int]
    routeFailedBetween: list[int]


class RouteStatisticsDict(TypedDict):
    """Represent a route statistics data dict type."""

    protocol_data_rate: int
    repeaters: list[Node]
    rssi: int | None
    repeater_rssi: list[int]
    route_failed_between: tuple[Node, Node] | None


@dataclass
class RouteStatistics:
    """Represent route statistics."""

    client: Client = field(repr=False)
    data: RouteStatisticsDataType = field(repr=False)
    protocol_data_rate: ProtocolDataRate = field(init=False)

    def __post_init__(self) -> None:
        """Post initialize."""
        self.protocol_data_rate = ProtocolDataRate(self.data["protocolDataRate"])

    @cached_property
    def repeaters(self) -> list[Node]:
        """Return repeaters."""
        assert self.client.driver
        return [
            self.client.driver.controller.nodes[int(node_id)]
            for node_id in self.data["repeaters"]
        ]

    @property
    def rssi(self) -> int | None:
        """Return RSSI."""
        if (rssi := self.data.get("rssi")) is None:
            return None
        if rssi in [item.value for item in RssiError]:
            raise RssiErrorReceived(RssiError(rssi))
        return rssi

    @property
    def repeater_rssi(self) -> list[int]:
        """Return repeater RSSI."""
        repeater_rssi = self.data.get("repeaterRSSI", [])
        rssi_errors = [item.value for item in RssiError]
        if any(rssi_ in rssi_errors for rssi_ in repeater_rssi):
            raise RepeaterRssiErrorReceived(repeater_rssi)

        return repeater_rssi

    @cached_property
    def route_failed_between(self) -> tuple[Node, Node] | None:
        """Return route failed between."""
        if (node_ids := self.data.get("routeFailedBetween")) is None:
            return None
        assert self.client.driver
        assert len(node_ids) == 2
        return (
            self.client.driver.controller.nodes[int(node_ids[0])],
            self.client.driver.controller.nodes[int(node_ids[1])],
        )

    def as_dict(self) -> RouteStatisticsDict:
        """Return route statistics as dict."""
        return {
            "protocol_data_rate": self.protocol_data_rate.value,
            "repeaters": self.repeaters,
            "rssi": self.data.get("rssi"),
            "repeater_rssi": self.data.get("repeaterRSSI", []),
            "route_failed_between": (
                (
                    self.route_failed_between[0],
                    self.route_failed_between[1],
                )
                if self.route_failed_between
                else None
            ),
        }



"""
Constants for lock related CCs.

Includes Door Lock and Lock CCs.
"""

from __future__ import annotations

from dataclasses import dataclass, field
from enum import Enum, IntEnum
from typing import TypedDict

from .. import CommandClass


class DoorLockMode(IntEnum):
    """Enum with all (known/used) Z-Wave lock states for CommandClass.DOOR_LOCK."""

    # https://github.com/zwave-js/node-zwave-js/blob/master/packages/zwave-js/src/lib/commandclass/DoorLockCC.ts#L56-L65
    UNSECURED = 0
    UNSECURED_WITH_TIMEOUT = 1
    INSIDE_UNSECURED = 2
    INSIDE_UNSECURED_WITH_TIMEOUT = 3
    OUTSIDE_UNSECURED = 4
    OUTSIDE_UNSECURED_WITH_TIMEOUT = 5
    UNKNOWN = 254
    SECURED = 255


class OperationType(IntEnum):
    """Enum with all (known/used) Z-Wave lock operation types."""

    # https://github.com/zwave-js/node-zwave-js/blob/master/packages/cc/src/lib/_Types.ts#L496
    CONSTANT = 1
    TIMED = 2


DOOR_LOCK_CC_UNSECURED_MAP = {
    OperationType.CONSTANT: {
        DoorLockMode.UNSECURED,
        DoorLockMode.INSIDE_UNSECURED,
        DoorLockMode.OUTSIDE_UNSECURED,
    },
    OperationType.TIMED: {
        DoorLockMode.UNSECURED_WITH_TIMEOUT,
        DoorLockMode.INSIDE_UNSECURED_WITH_TIMEOUT,
        DoorLockMode.OUTSIDE_UNSECURED_WITH_TIMEOUT,
    },
}


class LatchStatus(str, Enum):
    """Enum with all (known/used) Z-Wave latch statuses."""

    # https://github.com/zwave-js/node-zwave-js/blob/master/packages/cc/src/cc/DoorLockCC.ts#L854
    OPEN = "open"
    CLOSED = "closed"


class BoltStatus(str, Enum):
    """Enum with all (known/used) Z-Wave bolt statuses."""

    # https://github.com/zwave-js/node-zwave-js/blob/master/packages/cc/src/cc/DoorLockCC.ts#L854
    LOCKED = "locked"
    UNLOCKED = "unlocked"


class DoorStatus(str, Enum):
    """Enum with all (known/used) Z-Wave door statuses."""

    # https://github.com/zwave-js/node-zwave-js/blob/master/packages/cc/src/cc/DoorLockCC.ts#L854
    OPEN = "open"
    CLOSED = "closed"


class CodeSlotStatus(IntEnum):
    """Enum with all (known/used) Z-Wave code slot statuses."""

    AVAILABLE = 0
    ENABLED = 1
    DISABLED = 2


# Depending on the Command Class being used by the lock, the lock state is
# different so we need a map to track it
LOCK_CMD_CLASS_TO_LOCKED_STATE_MAP = {
    CommandClass.DOOR_LOCK: DoorLockMode.SECURED,
    CommandClass.LOCK: True,
}

# Door Lock CC value constants
BOLT_STATUS_PROPERTY = "boltStatus"
CURRENT_AUTO_RELOCK_TIME_PROPERTY = "autoRelockTime"
CURRENT_BLOCK_TO_BLOCK_PROPERTY = "blockToBlock"
CURRENT_HOLD_AND_RELEASE_TIME_PROPERTY = "holdAndReleaseTime"
CURRENT_INSIDE_HANDLES_CAN_OPEN_DOOR_PROPERTY = "insideHandlesCanOpenDoor"
CURRENT_LOCK_TIMEOUT_PROPERTY = "lockTimeout"
CURRENT_MODE_PROPERTY = "currentMode"
CURRENT_OPERATION_TYPE_PROPERTY = "operationType"
CURRENT_OUTSIDE_HANDLES_CAN_OPEN_DOOR_PROPERTY = "outsideHandlesCanOpenDoor"
CURRENT_TWIST_ASSIST_PROPERTY = "twistAssist"
DOOR_STATUS_PROPERTY = "doorStatus"
LATCH_STATUS_PROPERTY = "latchStatus"
TARGET_MODE_PROPERTY = "targetMode"

# Door Lock CC configuration constants
TARGET_AUTO_RELOCK_TIME_PROPERTY = CURRENT_AUTO_RELOCK_TIME_PROPERTY
TARGET_BLOCK_TO_BLOCK_PROPERTY = CURRENT_BLOCK_TO_BLOCK_PROPERTY
TARGET_HOLD_AND_RELEASE_TIME_PROPERTY = CURRENT_HOLD_AND_RELEASE_TIME_PROPERTY
TARGET_INSIDE_HANDLES_CAN_OPEN_DOOR_PROPERTY = "insideHandlesCanOpenDoorConfiguration"
TARGET_LOCK_TIMEOUT_PROPERTY = "lockTimeoutConfiguration"
TARGET_OPERATION_TYPE_PROPERTY = CURRENT_OPERATION_TYPE_PROPERTY
TARGET_OUTSIDE_HANDLES_CAN_OPEN_DOOR_PROPERTY = "outsideHandlesCanOpenDoorConfiguration"
TARGET_TWIST_ASSIST_PROPERTY = CURRENT_TWIST_ASSIST_PROPERTY

# Lock CC constants
LOCKED_PROPERTY = "locked"

# User Code CC constants
LOCK_USERCODE_PROPERTY = "userCode"
LOCK_USERCODE_ID_PROPERTY = "userId"
LOCK_USERCODE_STATUS_PROPERTY = "userIdStatus"

ATTR_CODE_SLOT = "code_slot"
ATTR_IN_USE = "in_use"
ATTR_NAME = "name"
ATTR_USERCODE = "usercode"

# Depending on the Command Class being used by the lock, the locked state property
# is different so we need a map to track it
LOCK_CMD_CLASS_TO_PROPERTY_MAP = {
    CommandClass.DOOR_LOCK: TARGET_MODE_PROPERTY,
    CommandClass.LOCK: LOCKED_PROPERTY,
}


class DoorLockCCConfigurationSetOptionsDataType(TypedDict, total=False):
    """Door Lock CC Configuration Set command options."""

    # https://github.com/zwave-js/node-zwave-js/blob/master/packages/cc/src/cc/DoorLockCC.ts#L1156
    operationType: int  # required
    lockTimeoutConfiguration: int
    outsideHandlesCanOpenDoorConfiguration: list[bool]  # required when from device
    insideHandlesCanOpenDoorConfiguration: list[bool]  # required when from device
    autoRelockTime: int
    holdAndReleaseTime: int
    twistAssist: bool
    blockToBlock: bool


DEFAULT_HANDLE_CONFIGURATION = [True, True, True, True]


@dataclass
class DoorLockCCConfigurationSetOptions:
    """Door Lock CC Configuration Set command options."""

    operation_type: OperationType
    lock_timeout_configuration: int | None = None
    outside_handles_can_open_door_configuration: list[bool] = field(
        default_factory=list
    )
    inside_handles_can_open_door_configuration: list[bool] = field(default_factory=list)
    auto_relock_time: int | None = None
    hold_and_release_time: int | None = None
    twist_assist: bool | None = None
    block_to_block: bool | None = None

    def __post_init__(self) -> None:
        """Post initialization."""
        for attr_name in (
            "inside_handles_can_open_door_configuration",
            "outside_handles_can_open_door_configuration",
        ):
            if not getattr(self, attr_name):
                setattr(self, attr_name, DEFAULT_HANDLE_CONFIGURATION)

    def to_dict(self) -> DoorLockCCConfigurationSetOptionsDataType:
        """Convert command options to dict."""
        data = DoorLockCCConfigurationSetOptionsDataType(
            operationType=self.operation_type,
            outsideHandlesCanOpenDoorConfiguration=(
                self.outside_handles_can_open_door_configuration
            ),
            insideHandlesCanOpenDoorConfiguration=(
                self.inside_handles_can_open_door_configuration
            ),
        )
        for prop_name, val in (
            (TARGET_LOCK_TIMEOUT_PROPERTY, self.lock_timeout_configuration),
            (TARGET_AUTO_RELOCK_TIME_PROPERTY, self.auto_relock_time),
            (TARGET_HOLD_AND_RELEASE_TIME_PROPERTY, self.hold_and_release_time),
            (TARGET_TWIST_ASSIST_PROPERTY, self.twist_assist),
            (TARGET_BLOCK_TO_BLOCK_PROPERTY, self.block_to_block),
        ):
            if val is not None:
                data[prop_name] = val  # type: ignore[literal-required]

        return data



"""Provide a model for the Z-Wave JS value."""

from __future__ import annotations

from dataclasses import dataclass, field
from enum import IntEnum, StrEnum
from typing import TYPE_CHECKING, Any, TypedDict

from ..const import (
    VALUE_UNKNOWN,
    CommandStatus,
    ConfigurationValueType,
    SetValueStatus,
    SupervisionStatus,
)
from ..event import Event
from ..util.helpers import parse_buffer
from .duration import Duration, DurationDataType

if TYPE_CHECKING:
    from .node import Node


class ValueType(StrEnum):
    """Enum with all value types."""

    ANY = "any"
    BOOLEAN = "boolean"
    NUMBER = "number"
    STRING = "string"


class MetaDataType(TypedDict, total=False):
    """Represent a metadata data dict type."""

    type: str  # required
    readable: bool  # required
    writeable: bool  # required
    description: str
    label: str
    min: int | None
    max: int | None
    unit: str | None
    states: dict[str, str]
    ccSpecific: dict[str, Any]
    valueChangeOptions: list[str]
    allowManualEntry: bool
    stateful: bool
    secret: bool
    default: int
    # Configuration Value specific attributes
    valueSize: int
    format: int
    noBulkSupport: bool  # deprecated
    isAdvanced: bool
    requiresReInclusion: bool
    isFromConfig: bool


class ValueDataType(TypedDict, total=False):
    """Represent a value data dict type."""

    commandClass: int  # required
    commandClassName: str  # required
    endpoint: int
    property: int | str  # required
    propertyName: str  # required
    propertyKey: int | str
    propertyKeyName: str
    value: Any
    newValue: Any
    prevValue: Any
    metadata: MetaDataType  # required
    ccVersion: int  # required


def _get_value_id_str_from_dict(node: Node, val: ValueDataType) -> str:
    """Return string ID of value from ValueDataType dict."""
    return get_value_id_str(
        node,
        val["commandClass"],
        val["property"],
        val.get("endpoint"),
        val.get("propertyKey"),
    )


def get_value_id_str(
    node: Node,
    command_class: int,
    property_: int | str,
    endpoint: int | None = None,
    property_key: int | str | None = None,
) -> str:
    """Return string ID of value."""
    # If endpoint is not provided, assume root endpoint
    endpoint_ = endpoint or 0
    value_id = f"{node.node_id}-{command_class}-{endpoint_}-{property_}"
    # Property key is only included when it has a value
    if property_key is not None:
        value_id += f"-{property_key}"
    return value_id


class ValueMetadata:
    """Represent metadata on a value instance."""

    def __init__(self, data: MetaDataType) -> None:
        """Initialize metadata."""
        self.data = data

    @property
    def type(self) -> str:
        """Return type."""
        return self.data["type"]

    @property
    def readable(self) -> bool | None:
        """Return readable."""
        return self.data.get("readable")

    @property
    def writeable(self) -> bool | None:
        """Return writeable."""
        return self.data.get("writeable")

    @property
    def label(self) -> str | None:
        """Return label."""
        return self.data.get("label")

    @property
    def description(self) -> str | None:
        """Return description."""
        return self.data.get("description")

    @property
    def min(self) -> int | None:
        """Return min."""
        return self.data.get("min")

    @property
    def max(self) -> int | None:
        """Return max."""
        return self.data.get("max")

    @property
    def unit(self) -> str | None:
        """Return unit."""
        return self.data.get("unit")

    @property
    def states(self) -> dict:
        """Return (optional) states."""
        return self.data.get("states", {})

    @property
    def cc_specific(self) -> dict[str, Any]:
        """Return ccSpecific."""
        return self.data.get("ccSpecific", {})

    @property
    def value_change_options(self) -> list[str]:
        """Return valueChangeOptions."""
        return self.data.get("valueChangeOptions", [])

    @property
    def allow_manual_entry(self) -> bool | None:
        """Return allowManualEntry."""
        return self.data.get("allowManualEntry")

    @property
    def value_size(self) -> int | None:
        """Return valueSize."""
        return self.data.get("valueSize")

    @property
    def stateful(self) -> bool | None:
        """Return stateful."""
        return self.data.get("stateful")

    @property
    def secret(self) -> bool | None:
        """Return secret."""
        return self.data.get("secret")

    @property
    def default(self) -> int | None:
        """Return default."""
        return self.data.get("default")

    @property
    def format(self) -> ConfigurationValueFormat | None:
        """Return format."""
        if (format_ := self.data.get("format")) is None:
            return None
        return ConfigurationValueFormat(format_)

    @property
    def no_bulk_support(self) -> bool | None:
        """Return noBulkSupport."""
        return self.data.get("noBulkSupport")

    @property
    def is_advanced(self) -> bool | None:
        """Return isAdvanced."""
        return self.data.get("isAdvanced")

    @property
    def requires_re_inclusion(self) -> bool | None:
        """Return requiresReInclusion."""
        return self.data.get("requiresReInclusion")

    @property
    def is_from_config(self) -> bool | None:
        """Return isFromConfig."""
        return self.data.get("isFromConfig")

    def update(self, data: MetaDataType) -> None:
        """Update data."""
        self.data.update(data)


class Value:
    """Represent a Z-Wave JS value."""

    def __init__(self, node: Node, data: ValueDataType) -> None:
        """Initialize value."""
        self.node = node
        self.data: ValueDataType = {}
        self._value: Any = None
        self._metadata = ValueMetadata({"type": "unknown"})
        self.update(data)

    def __repr__(self) -> str:
        """Return the representation."""
        return f"{type(self).__name__}(value_id={self.value_id!r})"

    def __hash__(self) -> int:
        """Return the hash."""
        return hash((self.node, self.value_id))

    def __eq__(self, other: object) -> bool:
        """Return whether this instance equals another."""
        if not isinstance(other, Value):
            return False
        return self.node == other.node and self.value_id == other.value_id

    @property
    def value_id(self) -> str:
        """Return value ID."""
        return _get_value_id_str_from_dict(self.node, self.data)

    @property
    def metadata(self) -> ValueMetadata:
        """Return value metadata."""
        return self._metadata

    @property
    def value(self) -> Any | None:
        """Return value."""
        # Treat unknown values like they are None
        if self._value == VALUE_UNKNOWN:
            return None
        return self._value

    @property
    def command_class_name(self) -> str:
        """Return commandClassName."""
        return self.data["commandClassName"]

    @property
    def command_class(self) -> int:
        """Return commandClass."""
        return self.data["commandClass"]

    @property
    def cc_version(self) -> int:
        """Return commandClass version."""
        return self.data["ccVersion"]

    @property
    def endpoint(self) -> int | None:
        """Return endpoint."""
        return self.data.get("endpoint")

    @property
    def property_(self) -> int | str:
        """Return property.

        Note the underscore in the end of this property name.
        That's there to not confuse Python to think it's a property
        decorator.
        """
        return self.data["property"]

    @property
    def property_key(self) -> int | str | None:
        """Return propertyKey."""
        return self.data.get("propertyKey")

    @property
    def property_name(self) -> str | None:
        """Return propertyName."""
        return self.data.get("propertyName")

    @property
    def property_key_name(self) -> str | None:
        """Return propertyKeyName."""
        return self.data.get("propertyKeyName")

    def receive_event(self, event: Event) -> None:
        """Receive an event."""
        self.update(event.data["args"])

    def update(self, data: ValueDataType) -> None:
        """Update data."""
        self.data.update(data)
        self.data.pop("prevValue", None)
        if "newValue" in self.data:
            self.data["value"] = self.data.pop("newValue")

        if "metadata" in data:
            self._metadata.update(data["metadata"])

        self._value = self.data.get("value")

        # Handle buffer dict and json string in value.
        if self._value is not None and self.metadata.type == "buffer":
            self._value = parse_buffer(self._value)


class ValueNotification(Value):
    """
    Model for a Value Notification message.

    https://zwave-js.github.io/node-zwave-js/#/api/node?id=quotvalue-notificationquot
    """

    # format is the same as a Value message, subclassed for easier identifying and
    # future use


class ConfigurationValueFormat(IntEnum):
    """Enum of all known configuration value formats."""

    # https://github.com/zwave-js/node-zwave-js/blob/cc_api_options/packages/core/src/values/Metadata.ts#L157
    SIGNED_INTEGER = 0
    UNSIGNED_INTEGER = 1
    ENUMERATED = 2
    BIT_FIELD = 3


class SupervisionResultDataType(TypedDict, total=False):
    """Represent a Supervision result data dict type."""

    # https://github.com/zwave-js/node-zwave-js/blob/cc_api_options/packages/core/src/consts/Transmission.ts#L311
    status: int
    remainingDuration: DurationDataType  # not included unless status is 1 (working)


@dataclass
class SupervisionResult:
    """Represent a Supervision result type."""

    data: SupervisionResultDataType = field(repr=False)
    status: SupervisionStatus = field(init=False)
    remaining_duration: Duration | None = field(init=False, default=None)

    def __post_init__(self) -> None:
        """Post initialization."""
        self.status = SupervisionStatus(self.data["status"])
        if remaining_duration := self.data.get("remainingDuration"):
            self.remaining_duration = Duration(remaining_duration)

        if self.status == SupervisionStatus.WORKING ^ bool(
            self.remaining_duration is not None
        ):
            raise ValueError(
                "SupervisionStatus of WORKING requires a remaining duration, all "
                "other statuses don't include it"
            )


class ConfigurationValue(Value):
    """Model for a Configuration Value."""

    @property
    def configuration_value_type(self) -> ConfigurationValueType:
        """Return configuration value type."""
        min_ = self.metadata.min
        max_ = self.metadata.max
        states = self.metadata.states
        allow_manual_entry = self.metadata.allow_manual_entry
        type_ = self.metadata.type

        if (max_ == 1 and min_ == 0 or type_ == ValueType.BOOLEAN) and not states:
            return ConfigurationValueType.BOOLEAN

        if (
            allow_manual_entry
            and not max_ == min_ == 0
            and not (max_ is None and min_ is None)
        ):
            return ConfigurationValueType.MANUAL_ENTRY

        if states:
            return ConfigurationValueType.ENUMERATED

        if (max_ is not None or min_ is not None) and not max_ == min_ == 0:
            return ConfigurationValueType.RANGE

        return ConfigurationValueType.UNDEFINED


class SetValueResultDataType(TypedDict, total=False):
    """Represent a setValue result data dict type."""

    # https://github.com/zwave-js/node-zwave-js/blob/v11-dev/packages/cc/src/lib/API.ts#L103
    status: int  # required
    remainingDuration: DurationDataType
    message: str


@dataclass
class SetValueResult:
    """Result from setValue command."""

    data: SetValueResultDataType = field(repr=False)
    status: SetValueStatus = field(init=False)
    remaining_duration: Duration | None = field(init=False)
    message: str | None = field(init=False)

    def __post_init__(self) -> None:
        """Post init."""
        self.status = SetValueStatus(self.data["status"])
        self.remaining_duration = (
            Duration(duration_data)
            if (duration_data := self.data.get("remainingDuration"))
            else None
        )
        self.message = self.data.get("message")

    def __repr__(self) -> str:
        """Return the representation."""
        status = self.status.name.replace("_", " ").title()
        if self.status == SetValueStatus.WORKING:
            assert self.remaining_duration
            return f"{status} ({self.remaining_duration})"
        if self.status in (
            SetValueStatus.ENDPOINT_NOT_FOUND,
            SetValueStatus.INVALID_VALUE,
            SetValueStatus.NOT_IMPLEMENTED,
        ):
            assert self.message
            return f"{status}: {self.message}"
        return status


@dataclass
class SetConfigParameterResult:
    """Result of a set config parameter command."""

    status: CommandStatus
    result: SupervisionResult | SetValueResult | None = None



"""Provide a model for the Z-Wave JS node's events."""

from __future__ import annotations

from typing import Literal

from pydantic import BaseModel

from ...const import CommandClass
from ...event import BaseEventModel
from ..notification import (
    EntryControlNotificationArgsDataType,
    MultilevelSwitchNotificationArgsDataType,
    NotificationNotificationArgsDataType,
    PowerLevelNotificationArgsDataType,
)
from ..value import ValueDataType
from .data_model import NodeDataType
from .firmware import (
    NodeFirmwareUpdateProgressDataType,
    NodeFirmwareUpdateResultDataType,
)
from .statistics import NodeStatisticsDataType


class BaseNodeEventModel(BaseEventModel):
    """Base model for a node event."""

    source: Literal["node"]
    nodeId: int

    @classmethod
    def from_dict(cls, data: dict) -> BaseNodeEventModel:
        """Initialize from dict."""
        return cls(
            source=data["source"],
            event=data["event"],
            nodeId=data["nodeId"],
        )


class AliveEventModel(BaseNodeEventModel):
    """Model for `alive` event data."""

    event: Literal["alive"]


class CheckHealthProgressEventModel(BaseNodeEventModel):
    """
    Model for `check health progress` type events data.

    Includes `check lifeline health progress` and `check route health progress` events.
    """

    rounds: int
    totalRounds: int
    lastRating: int

    @classmethod
    def from_dict(cls, data: dict) -> CheckHealthProgressEventModel:
        """Initialize from dict."""
        return cls(
            source=data["source"],
            event=data["event"],
            nodeId=data["nodeId"],
            rounds=data["rounds"],
            totalRounds=data["totalRounds"],
            lastRating=data["lastRating"],
        )


class CheckLifelineHealthProgressEventModel(CheckHealthProgressEventModel):
    """Model for `check lifeline health progress` event data."""

    event: Literal["check lifeline health progress"]


class CheckRouteHealthProgressEventModel(CheckHealthProgressEventModel):
    """Model for `check route health progress` event data."""

    event: Literal["check route health progress"]


class DeadEventModel(BaseNodeEventModel):
    """Model for `dead` event data."""

    event: Literal["dead"]


class InterviewCompletedEventModel(BaseNodeEventModel):
    """Model for `interview completed` event data."""

    event: Literal["interview completed"]


class InterviewFailedEventArgsModel(BaseModel):
    """Model for `interview failed` event args."""

    errorMessage: str
    isFinal: bool
    attempt: int | None = None
    maxAttempts: int | None = None

    @classmethod
    def from_dict(cls, data: dict) -> InterviewFailedEventArgsModel:
        """Initialize from dict."""
        return cls(
            source=data["source"],
            event=data["event"],
            nodeId=data["nodeId"],
            errorMessage=data["errorMessage"],
            isFinal=data["isFinal"],
            attempt=data["attempt"],
            maxAttempts=data["maxAttempts"],
        )


class InterviewFailedEventModel(BaseNodeEventModel):
    """Model for `interview failed` event data."""

    event: Literal["interview failed"]
    args: InterviewFailedEventArgsModel

    @classmethod
    def from_dict(cls, data: dict) -> InterviewFailedEventModel:
        """Initialize from dict."""
        return cls(
            source=data["source"],
            event=data["event"],
            nodeId=data["nodeId"],
            args=data["args"],
        )


class InterviewStageCompletedEventModel(BaseNodeEventModel):
    """Model for `interview stage completed` event data."""

    event: Literal["interview stage completed"]
    stageName: str

    @classmethod
    def from_dict(cls, data: dict) -> InterviewStageCompletedEventModel:
        """Initialize from dict."""
        return cls(
            source=data["source"],
            event=data["event"],
            nodeId=data["nodeId"],
            stageName=data["stageName"],
        )


class InterviewStartedEventModel(BaseNodeEventModel):
    """Model for `interview started` event data."""

    event: Literal["interview started"]


class NotificationEventModel(BaseNodeEventModel):
    """Model for `notification` event data."""

    event: Literal["notification"]
    nodeId: int
    endpointIndex: int
    ccId: CommandClass
    args: (
        NotificationNotificationArgsDataType
        | EntryControlNotificationArgsDataType
        | PowerLevelNotificationArgsDataType
        | MultilevelSwitchNotificationArgsDataType
    )

    @classmethod
    def from_dict(cls, data: dict) -> NotificationEventModel:
        """Initialize from dict."""
        return cls(
            source=data["source"],
            event=data["event"],
            nodeId=data["nodeId"],
            endpointIndex=data["endpointIndex"],
            ccId=data["ccId"],
            args=data["args"],
        )


class ReadyEventModel(BaseNodeEventModel):
    """Model for `ready` event data."""

    event: Literal["ready"]
    nodeState: NodeDataType

    @classmethod
    def from_dict(cls, data: dict) -> ReadyEventModel:
        """Initialize from dict."""
        return cls(
            source=data["source"],
            event=data["event"],
            nodeId=data["nodeId"],
            nodeState=data["nodeState"],
        )


class SleepEventModel(BaseNodeEventModel):
    """Model for `sleep` event data."""

    event: Literal["sleep"]


class StatisticsUpdatedEventModel(BaseNodeEventModel):
    """Model for `statistics updated` event data."""

    event: Literal["statistics updated"]
    statistics: NodeStatisticsDataType

    @classmethod
    def from_dict(cls, data: dict) -> StatisticsUpdatedEventModel:
        """Initialize from dict."""
        return cls(
            source=data["source"],
            event=data["event"],
            nodeId=data["nodeId"],
            statistics=data["statistics"],
        )


class TestPowerLevelProgressEventModel(BaseNodeEventModel):
    """Model for `test powerlevel progress` event data."""

    event: Literal["test powerlevel progress"]
    acknowledged: int
    total: int

    @classmethod
    def from_dict(cls, data: dict) -> TestPowerLevelProgressEventModel:
        """Initialize from dict."""
        return cls(
            source=data["source"],
            event=data["event"],
            nodeId=data["nodeId"],
            acknowledged=data["acknowledged"],
            total=data["total"],
        )


class ValueEventModel(BaseNodeEventModel):
    """
    Model for `value` events data.

    Subclass for event models for `metadata updated`, `value added`,
    `value notification`, `value removed`, and `value updated`.
    """

    args: ValueDataType

    @classmethod
    def from_dict(cls, data: dict) -> ValueEventModel:
        """Initialize from dict."""
        return cls(
            source=data["source"],
            event=data["event"],
            nodeId=data["nodeId"],
            args=data["args"],
        )


class MetadataUpdatedEventModel(ValueEventModel):
    """Model for `metadata updated` event data."""

    event: Literal["metadata updated"]


class ValueAddedEventModel(ValueEventModel):
    """Model for `value added` event data."""

    event: Literal["value added"]


class ValueNotificationEventModel(ValueEventModel):
    """Model for `value notification` event data."""

    event: Literal["value notification"]


class ValueRemovedEventModel(ValueEventModel):
    """Model for `value removed` event data."""

    event: Literal["value removed"]


class ValueUpdatedEventModel(ValueEventModel):
    """Model for `value updated` event data."""

    event: Literal["value updated"]


class WakeUpEventModel(BaseNodeEventModel):
    """Model for `wake up` event data."""

    event: Literal["wake up"]


class FirmwareUpdateFinishedEventModel(BaseNodeEventModel):
    """Model for `firmware update finished` event data."""

    event: Literal["firmware update finished"]
    result: NodeFirmwareUpdateResultDataType

    @classmethod
    def from_dict(cls, data: dict) -> FirmwareUpdateFinishedEventModel:
        """Initialize from dict."""
        return cls(
            source=data["source"],
            event=data["event"],
            nodeId=data["nodeId"],
            result=data["result"],
        )


class FirmwareUpdateProgressEventModel(BaseNodeEventModel):
    """Model for `firmware update progress` event data."""

    event: Literal["firmware update progress"]
    progress: NodeFirmwareUpdateProgressDataType

    @classmethod
    def from_dict(cls, data: dict) -> FirmwareUpdateProgressEventModel:
        """Initialize from dict."""
        return cls(
            source=data["source"],
            event=data["event"],
            nodeId=data["nodeId"],
            progress=data["progress"],
        )


NODE_EVENT_MODEL_MAP: dict[str, type[BaseNodeEventModel]] = {
    "alive": AliveEventModel,
    "check lifeline health progress": CheckLifelineHealthProgressEventModel,
    "check route health progress": CheckRouteHealthProgressEventModel,
    "dead": DeadEventModel,
    "firmware update finished": FirmwareUpdateFinishedEventModel,
    "firmware update progress": FirmwareUpdateProgressEventModel,
    "interview completed": InterviewCompletedEventModel,
    "interview failed": InterviewFailedEventModel,
    "interview stage completed": InterviewStageCompletedEventModel,
    "interview started": InterviewStartedEventModel,
    "metadata updated": MetadataUpdatedEventModel,
    "notification": NotificationEventModel,
    "ready": ReadyEventModel,
    "sleep": SleepEventModel,
    "statistics updated": StatisticsUpdatedEventModel,
    "test powerlevel progress": TestPowerLevelProgressEventModel,
    "value added": ValueAddedEventModel,
    "value notification": ValueNotificationEventModel,
    "value removed": ValueRemovedEventModel,
    "value updated": ValueUpdatedEventModel,
    "wake up": WakeUpEventModel,
}



"""Provide a model for Z-Wave firmware."""

from __future__ import annotations

from dataclasses import dataclass, field
from enum import IntEnum
from typing import TYPE_CHECKING, Required, TypedDict, cast

from zwave_js_server.model.firmware import (
    FirmwareUpdateData,
    FirmwareUpdateDataDataType,
    FirmwareUpdateDeviceID,
    FirmwareUpdateDeviceIDDataType,
    FirmwareUpdateFileInfo,
    FirmwareUpdateFileInfoDataType,
    FirmwareUpdateInfo,
    FirmwareUpdateInfoDataType,
    FirmwareUpdateProgress,
    FirmwareUpdateProgressDataType,
    FirmwareUpdateResult,
    FirmwareUpdateResultDataType,
)

from ...const import VALUE_UNKNOWN

if TYPE_CHECKING:
    from . import Node


class NodeFirmwareUpdateDataDataType(FirmwareUpdateDataDataType):
    """Represent a firmware update data dict type."""

    firmwareTarget: int


@dataclass
class NodeFirmwareUpdateData(FirmwareUpdateData):
    """Firmware update data."""

    firmware_target: int | None = None

    def to_dict(self) -> NodeFirmwareUpdateDataDataType:
        """Convert firmware update data to dict."""
        data = super().to_dict()
        node_data = cast(NodeFirmwareUpdateDataDataType, data)
        if self.firmware_target is not None:
            node_data["firmwareTarget"] = self.firmware_target
        return node_data


class NodeFirmwareUpdateCapabilitiesDataType(TypedDict, total=False):
    """Represent a firmware update capabilities dict type."""

    firmwareUpgradable: Required[bool]
    firmwareTargets: list[int]
    continuesToFunction: bool | str
    supportsActivation: bool | str


class NodeFirmwareUpdateCapabilitiesDict(TypedDict, total=False):
    """Represent a dict from FirmwareUpdateCapabilities."""

    firmware_upgradable: Required[bool]
    firmware_targets: list[int]
    continues_to_function: bool | None
    supports_activation: bool | None


@dataclass
class NodeFirmwareUpdateCapabilities:
    """Model for firmware update capabilities."""

    data: NodeFirmwareUpdateCapabilitiesDataType = field(repr=False)
    firmware_upgradable: bool = field(init=False)

    def __post_init__(self) -> None:
        """Post initialize."""
        self.firmware_upgradable = self.data["firmwareUpgradable"]

    @property
    def firmware_targets(self) -> list[int]:
        """Return firmware targets."""
        if not self.firmware_upgradable:
            raise TypeError("Firmware is not upgradeable.")
        return self.data["firmwareTargets"]

    @property
    def continues_to_function(self) -> bool | None:
        """Return whether node continues to function during update."""
        if not self.firmware_upgradable:
            raise TypeError("Firmware is not upgradeable.")
        if (val := self.data["continuesToFunction"]) == VALUE_UNKNOWN:
            return None
        assert isinstance(val, bool)
        return val

    @property
    def supports_activation(self) -> bool | None:
        """Return whether node supports delayed activation of the new firmware."""
        if not self.firmware_upgradable:
            raise TypeError("Firmware is not upgradeable.")
        if (val := self.data["supportsActivation"]) == VALUE_UNKNOWN:
            return None
        assert isinstance(val, bool)
        return val

    def to_dict(self) -> NodeFirmwareUpdateCapabilitiesDict:
        """Return dict representation of the object."""
        if not self.firmware_upgradable:
            return {"firmware_upgradable": self.firmware_upgradable}
        return {
            "firmware_upgradable": self.firmware_upgradable,
            "firmware_targets": self.firmware_targets,
            "continues_to_function": self.continues_to_function,
            "supports_activation": self.supports_activation,
        }


class NodeFirmwareUpdateStatus(IntEnum):
    """Enum with all node firmware update status values.

    https://zwave-js.github.io/node-zwave-js/#/api/node?id=quotfirmware-update-finishedquot
    """

    ERROR_TIMEOUT = -1
    ERROR_CHECKSUM = 0
    ERROR_TRANSMISSION_FAILED = 1
    ERROR_INVALID_MANUFACTURER_ID = 2
    ERROR_INVALID_FIRMWARE_ID = 3
    ERROR_INVALID_FIRMWARE_TARGET = 4
    ERROR_INVALID_HEADER_INFORMATION = 5
    ERROR_INVALID_HEADER_FORMAT = 6
    ERROR_INSUFFICIENT_MEMORY = 7
    ERROR_INVALID_HARDWARE_VERSION = 8
    OK_WAITING_FOR_ACTIVATION = 253
    OK_NO_RESTART = 254
    OK_RESTART_PENDING = 255


class NodeFirmwareUpdateProgressDataType(FirmwareUpdateProgressDataType):
    """Represent a node firmware update progress dict type."""

    currentFile: int
    totalFiles: int


@dataclass
class NodeFirmwareUpdateProgress(FirmwareUpdateProgress):
    """Model for a node firmware update progress data."""

    node: Node
    data: NodeFirmwareUpdateProgressDataType = field(repr=False)
    current_file: int = field(init=False)
    total_files: int = field(init=False)

    def __init__(self, node: Node, data: NodeFirmwareUpdateProgressDataType) -> None:
        """Initialize the node firmware update progress.

        Explicit init method to retain backwards compatibility
        that requires the node as first parameter.
        """
        super().__init__(data)
        self.node = node

    def __post_init__(self) -> None:
        """Post initialize."""
        super().__post_init__()
        self.current_file = self.data["currentFile"]
        self.total_files = self.data["totalFiles"]


class NodeFirmwareUpdateResultDataType(FirmwareUpdateResultDataType, total=False):
    """Represent a node firmware update result dict type."""

    waitTime: int
    reInterview: Required[bool]


@dataclass
class NodeFirmwareUpdateResult(FirmwareUpdateResult):
    """Model for node firmware update result data."""

    node: Node
    data: NodeFirmwareUpdateResultDataType = field(repr=False)
    status: NodeFirmwareUpdateStatus = field(init=False)
    wait_time: int | None = field(init=False)
    reinterview: bool = field(init=False)
    _status_class = NodeFirmwareUpdateStatus

    def __init__(self, node: Node, data: NodeFirmwareUpdateResultDataType) -> None:
        """Initialize the node firmware update result.

        Explicit init method to retain backwards compatibility
        that requires the node as first parameter.
        """
        super().__init__(data)
        self.node = node

    def __post_init__(self) -> None:
        """Post initialize."""
        super().__post_init__()
        self.wait_time = self.data.get("waitTime")
        self.reinterview = self.data["reInterview"]


class NodeFirmwareUpdateFileInfoDataType(FirmwareUpdateFileInfoDataType):
    """Represent a node firmware update file info data dict type."""


@dataclass
class NodeFirmwareUpdateFileInfo(FirmwareUpdateFileInfo):
    """Represent a firmware update file info."""

    @classmethod
    def from_dict(
        cls, data: NodeFirmwareUpdateFileInfoDataType
    ) -> NodeFirmwareUpdateFileInfo:
        """Initialize from dict."""
        return cast(NodeFirmwareUpdateFileInfo, super().from_dict(data))

    def to_dict(self) -> NodeFirmwareUpdateFileInfoDataType:
        """Return dict representation of the object."""
        return super().to_dict()


class NodeFirmwareUpdateDeviceIDDataType(FirmwareUpdateDeviceIDDataType):
    """Represent a node firmware update device ID data dict type."""


@dataclass
class NodeFirmwareUpdateDeviceID(FirmwareUpdateDeviceID):
    """Represent a firmware update device ID."""

    @classmethod
    def from_dict(
        cls, data: NodeFirmwareUpdateDeviceIDDataType
    ) -> NodeFirmwareUpdateDeviceID:
        """Initialize from dict."""
        return cast(NodeFirmwareUpdateDeviceID, super().from_dict(data))

    def to_dict(self) -> NodeFirmwareUpdateDeviceIDDataType:
        """Return dict representation of the object."""
        return super().to_dict()


class NodeFirmwareUpdateInfoDataType(FirmwareUpdateInfoDataType):
    """Represent a node firmware update info data dict type."""


@dataclass
class NodeFirmwareUpdateInfo(FirmwareUpdateInfo):
    """Represent a firmware update info."""

    @classmethod
    def from_dict(cls, data: NodeFirmwareUpdateInfoDataType) -> NodeFirmwareUpdateInfo:
        """Initialize from dict."""
        return cast(NodeFirmwareUpdateInfo, super().from_dict(data))

    def to_dict(self) -> NodeFirmwareUpdateInfoDataType:
        """Return dict representation of the object."""
        return super().to_dict()



"""Provide models for Z-Wave JS firmware."""

from __future__ import annotations

from dataclasses import asdict, dataclass, field
from enum import IntEnum
from typing import Literal, Required, TypedDict, cast

from zwave_js_server.const import RFRegion
from zwave_js_server.util.helpers import convert_bytes_to_base64


@dataclass
class FirmwareUpdateData:
    """Firmware update data."""

    filename: str
    file: bytes
    file_format: str | None = None

    def to_dict(self) -> FirmwareUpdateDataDataType:
        """Convert firmware update data to dict."""
        data: FirmwareUpdateDataDataType = {
            "filename": self.filename,
            "file": convert_bytes_to_base64(self.file),
        }
        if self.file_format is not None:
            data["fileFormat"] = self.file_format
        return data


class FirmwareUpdateDataDataType(TypedDict, total=False):
    """Represent a firmware update data dict type."""

    filename: Required[str]
    file: Required[str]
    fileFormat: str


@dataclass
class FirmwareUpdateInfo:
    """Represent a firmware update info."""

    version: str
    changelog: str
    channel: Literal["stable", "beta"]
    files: list[FirmwareUpdateFileInfo]
    downgrade: bool
    normalized_version: str
    device: FirmwareUpdateDeviceID

    @classmethod
    def from_dict(cls, data: FirmwareUpdateInfoDataType) -> FirmwareUpdateInfo:
        """Initialize from dict."""
        return cls(
            version=data["version"],
            changelog=data["changelog"],
            channel=data["channel"],
            files=[FirmwareUpdateFileInfo.from_dict(file) for file in data["files"]],
            downgrade=data["downgrade"],
            normalized_version=data["normalizedVersion"],
            device=FirmwareUpdateDeviceID.from_dict(data["device"]),
        )

    def to_dict(self) -> FirmwareUpdateInfoDataType:
        """Return dict representation of the object."""
        return cast(
            FirmwareUpdateInfoDataType,
            {
                "version": self.version,
                "changelog": self.changelog,
                "channel": self.channel,
                "files": [file.to_dict() for file in self.files],
                "downgrade": self.downgrade,
                "normalizedVersion": self.normalized_version,
                "device": self.device.to_dict(),
            },
        )


class FirmwareUpdateInfoDataType(TypedDict, total=False):
    """Represent a firmware update info data dict type."""

    version: str
    changelog: str
    channel: Literal["stable", "beta"]
    files: list[FirmwareUpdateFileInfoDataType]
    downgrade: bool
    normalizedVersion: str
    device: FirmwareUpdateDeviceIDDataType


@dataclass
class FirmwareUpdateDeviceID:
    """Represent a firmware update device ID."""

    manufacturer_id: int
    product_type: int
    product_id: int
    firmware_version: str
    rf_region: RFRegion | None

    @classmethod
    def from_dict(cls, data: FirmwareUpdateDeviceIDDataType) -> FirmwareUpdateDeviceID:
        """Initialize from dict."""
        return cls(
            manufacturer_id=data["manufacturerId"],
            product_type=data["productType"],
            product_id=data["productId"],
            firmware_version=data["firmwareVersion"],
            rf_region=RFRegion(data["rfRegion"]) if "rfRegion" in data else None,
        )

    def to_dict(self) -> FirmwareUpdateDeviceIDDataType:
        """Return dict representation of the object."""
        data = {
            "manufacturerId": self.manufacturer_id,
            "productType": self.product_type,
            "productId": self.product_id,
            "firmwareVersion": self.firmware_version,
        }
        if self.rf_region is not None:
            data["rfRegion"] = self.rf_region
        return cast(FirmwareUpdateDeviceIDDataType, data)


class FirmwareUpdateDeviceIDDataType(TypedDict, total=False):
    """Represent a firmware update device ID dict type."""

    manufacturerId: Required[int]
    productType: Required[int]
    productId: Required[int]
    firmwareVersion: Required[str]
    rfRegion: int


@dataclass
class FirmwareUpdateFileInfo:
    """Represent a firmware update file info."""

    target: int
    url: str
    integrity: str

    @classmethod
    def from_dict(cls, data: FirmwareUpdateFileInfoDataType) -> FirmwareUpdateFileInfo:
        """Initialize from dict."""
        return cls(
            target=data["target"],
            url=data["url"],
            integrity=data["integrity"],
        )

    def to_dict(self) -> FirmwareUpdateFileInfoDataType:
        """Return dict representation of the object."""
        return cast(FirmwareUpdateFileInfoDataType, asdict(self))


class FirmwareUpdateFileInfoDataType(TypedDict):
    """Represent a firmware update file info data dict type."""

    target: int
    url: str
    integrity: str  # sha256


@dataclass
class FirmwareUpdateProgress:
    """Model for a firmware update progress."""

    data: FirmwareUpdateProgressDataType = field(repr=False)
    sent_fragments: int = field(init=False)
    total_fragments: int = field(init=False)
    progress: float = field(init=False)

    def __post_init__(self) -> None:
        """Post initialize."""
        self.sent_fragments = self.data["sentFragments"]
        self.total_fragments = self.data["totalFragments"]
        self.progress = float(self.data["progress"])


class FirmwareUpdateProgressDataType(TypedDict):
    """Represent a firmware update progress dict type."""

    sentFragments: int
    totalFragments: int
    progress: float


@dataclass
class FirmwareUpdateResult:
    """Model for firmware update result data."""

    data: FirmwareUpdateResultDataType = field(repr=False)
    status: IntEnum = field(init=False)
    success: bool = field(init=False)
    _status_class: type[IntEnum] = field(init=False, repr=False)

    def __post_init__(self) -> None:
        """Post initialize."""
        self.status = self._status_class(self.data["status"])
        self.success = self.data["success"]


class FirmwareUpdateResultDataType(TypedDict):
    """Represent a driver firmware update result dict type."""

    status: int
    success: bool



"""Constants for Meter CC."""

from __future__ import annotations

from enum import IntEnum

VALUE_PROPERTY = "value"
RESET_PROPERTY = "reset"

CC_SPECIFIC_SCALE = "scale"
CC_SPECIFIC_METER_TYPE = "meterType"
CC_SPECIFIC_RATE_TYPE = "rateType"

RESET_METER_CC_API = "reset"

# optional attributes when calling the Meter CC reset API.
# https://github.com/zwave-js/node-zwave-js/blob/master/packages/cc/src/cc/MeterCC.ts
RESET_METER_OPTION_TARGET_VALUE = "targetValue"
RESET_METER_OPTION_TYPE = "type"
RESET_METER_OPTION_SCALE = "scale"
RESET_METER_OPTION_RATE_TYPE = "rateType"


# https://github.com/zwave-js/node-zwave-js/blob/master/packages/config/config/meters.json
class MeterType(IntEnum):
    """Enum with all known meter types."""

    ELECTRIC = 1
    GAS = 2
    WATER = 3
    HEATING = 4
    COOLING = 5


class RateType(IntEnum):
    """Enum with all known rate types."""

    # https://github.com/raman325/node-zwave-js/blob/master/packages/cc/src/lib/_Types.ts#L996
    UNSPECIFIED = 0
    CONSUMED = 1
    PRODUCED = 2


class MeterScaleType(IntEnum):
    """Common base class for meter scale enums."""


class ElectricScale(MeterScaleType):
    """Enum with all known electric meter scale values."""

    KILOWATT_HOUR = 0
    KILOVOLT_AMPERE_HOUR = 1
    WATT = 2
    PULSE_COUNT = 3
    VOLT = 4
    AMPERE = 5
    POWER_FACTOR = 6
    KILOVOLT_AMPERE_REACTIVE = 7
    KILOVOLT_AMPERE_REACTIVE_HOUR = 8


class GasScale(MeterScaleType):
    """Enum with all known gas meter scale values."""

    CUBIC_METER = 0
    CUBIC_FEET = 1
    PULSE_COUNT = 3


class WaterScale(MeterScaleType):
    """Enum with all known water meter scale values."""

    CUBIC_METER = 0
    CUBIC_FEET = 1
    US_GALLON = 2
    PULSE_COUNT = 3


class HeatingScale(MeterScaleType):
    """Enum with all known heating meter scale values."""

    KILOWATT_HOUR = 0


CoolingScale = HeatingScale

METER_TYPE_TO_SCALE_ENUM_MAP: dict[MeterType, type[MeterScaleType]] = {
    MeterType.ELECTRIC: ElectricScale,
    MeterType.GAS: GasScale,
    MeterType.WATER: WaterScale,
    MeterType.HEATING: HeatingScale,
    MeterType.COOLING: CoolingScale,
}

ENERGY_TOTAL_INCREASING_METER_TYPES: list[MeterScaleType] = [
    ElectricScale.KILOWATT_HOUR,
    ElectricScale.KILOVOLT_AMPERE_HOUR,
    ElectricScale.KILOVOLT_AMPERE_REACTIVE_HOUR,
    HeatingScale.KILOWATT_HOUR,
    CoolingScale.KILOWATT_HOUR,
    ElectricScale.PULSE_COUNT,
]
POWER_METER_TYPES: list[MeterScaleType] = [
    ElectricScale.WATT,
    ElectricScale.KILOVOLT_AMPERE_REACTIVE,
]
POWER_FACTOR_METER_TYPES: list[MeterScaleType] = [ElectricScale.POWER_FACTOR]
VOLTAGE_METER_TYPES: list[MeterScaleType] = [ElectricScale.VOLT]
CURRENT_METER_TYPES: list[MeterScaleType] = [ElectricScale.AMPERE]
GAS_METER_TYPES: list[MeterScaleType] = [
    GasScale.CUBIC_METER,
    GasScale.CUBIC_FEET,
    GasScale.PULSE_COUNT,
]
WATER_METER_TYPES: list[MeterScaleType] = [
    WaterScale.CUBIC_METER,
    WaterScale.CUBIC_FEET,
    WaterScale.US_GALLON,
    WaterScale.PULSE_COUNT,
]

UNIT_KILOWATT_HOUR: list[MeterScaleType] = [
    ElectricScale.KILOWATT_HOUR,
    HeatingScale.KILOWATT_HOUR,
    CoolingScale.KILOWATT_HOUR,
]
UNIT_KILOVOLT_AMPERE_HOUR: list[MeterScaleType] = [ElectricScale.KILOVOLT_AMPERE_HOUR]
UNIT_WATT: list[MeterScaleType] = [ElectricScale.WATT]
UNIT_PULSE_COUNT: list[MeterScaleType] = [
    ElectricScale.PULSE_COUNT,
    GasScale.PULSE_COUNT,
    WaterScale.PULSE_COUNT,
]
UNIT_VOLT: list[MeterScaleType] = [ElectricScale.VOLT]
UNIT_AMPERE: list[MeterScaleType] = [ElectricScale.AMPERE]
UNIT_POWER_FACTOR: list[MeterScaleType] = [ElectricScale.POWER_FACTOR]
UNIT_KILOVOLT_AMPERE_REACTIVE: list[MeterScaleType] = [
    ElectricScale.KILOVOLT_AMPERE_REACTIVE
]
UNIT_KILOVOLT_AMPERE_REACTIVE_HOUR: list[MeterScaleType] = [
    ElectricScale.KILOVOLT_AMPERE_REACTIVE_HOUR
]
UNIT_CUBIC_METER: list[MeterScaleType] = [GasScale.CUBIC_METER, WaterScale.CUBIC_METER]
UNIT_CUBIC_FEET: list[MeterScaleType] = [GasScale.CUBIC_FEET, WaterScale.CUBIC_FEET]
UNIT_US_GALLON: list[MeterScaleType] = [WaterScale.US_GALLON]



"""
Model for a Zwave Node's Notification Event.

https://zwave-js.github.io/node-zwave-js/#/api/node?id=quotnotificationquot
"""

from __future__ import annotations

from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Any, Literal, TypedDict

from ..const.command_class.multilevel_switch import MultilevelSwitchCommand
from ..const.command_class.power_level import PowerLevelTestStatus
from ..util.helpers import parse_buffer

if TYPE_CHECKING:
    from .node import Node


class BaseNotificationDataType(TypedDict):
    """Represent a base notification event data dict type."""

    source: Literal["node"]  # required
    event: Literal["notification"]  # required
    nodeId: int  # required
    endpointIndex: int  # required
    ccId: int  # required


@dataclass
class BaseNotification:
    """Model for a Zwave Node's notification event."""

    node: Node
    data: BaseNotificationDataType = field(repr=False)
    node_id: int = field(init=False, repr=False)
    endpoint_idx: int = field(init=False)
    command_class: int = field(init=False)

    def __post_init__(self) -> None:
        """Post initialization."""
        self.node_id = self.data["nodeId"]
        self.endpoint_idx = self.data["endpointIndex"]
        self.command_class = self.data["ccId"]


class EntryControlNotificationArgsDataType(TypedDict, total=False):
    """Represent args for a Entry Control CC notification event data dict type."""

    eventType: int  # required
    eventTypeLabel: str  # required
    dataType: int  # required
    dataTypeLabel: str  # required
    eventData: str | dict[str, Any]


class EntryControlNotificationDataType(BaseNotificationDataType):
    """Represent an Entry Control CC notification event data dict type."""

    args: EntryControlNotificationArgsDataType  # required


@dataclass
class EntryControlNotification(BaseNotification):
    """Model for a Zwave Node's Entry Control CC notification event."""

    data: EntryControlNotificationDataType = field(repr=False)
    event_type: int = field(init=False)
    event_type_label: str = field(init=False)
    data_type: int = field(init=False)
    data_type_label: str = field(init=False)
    event_data: str | dict[str, Any] | None = field(init=False, default=None)

    def __post_init__(self) -> None:
        """Post initialize."""
        super().__post_init__()
        self.event_type = self.data["args"]["eventType"]
        self.event_type_label = self.data["args"]["eventTypeLabel"]
        self.data_type = self.data["args"]["dataType"]
        self.data_type_label = self.data["args"]["dataTypeLabel"]
        if event_data := self.data["args"].get("eventData"):
            self.event_data = parse_buffer(event_data)


class NotificationNotificationArgsDataType(TypedDict, total=False):
    """Represent args for a Notification CC notification event data dict type."""

    type: int  # required
    label: str  # required
    event: int  # required
    eventLabel: str  # required
    parameters: dict[str, Any]


class NotificationNotificationDataType(BaseNotificationDataType):
    """Represent a Notification CC notification event data dict type."""

    args: NotificationNotificationArgsDataType  # required


@dataclass
class NotificationNotification(BaseNotification):
    """Model for a Zwave Node's Notification CC notification event."""

    data: NotificationNotificationDataType = field(repr=False)
    type_: int = field(init=False)
    label: str = field(init=False)
    event: int = field(init=False)
    event_label: str = field(init=False)
    parameters: dict[str, Any] = field(init=False)

    def __post_init__(self) -> None:
        """Post initialize."""
        super().__post_init__()
        self.type_ = self.data["args"]["type"]
        self.label = self.data["args"]["label"]
        self.event = self.data["args"]["event"]
        self.event_label = self.data["args"]["eventLabel"]
        self.parameters = self.data["args"].get("parameters", {})


class PowerLevelNotificationArgsDataType(TypedDict):
    """Represent args for a Power Level CC notification event data dict type."""

    testNodeId: int
    status: int
    acknowledgedFrames: int


class PowerLevelNotificationDataType(BaseNotificationDataType):
    """Represent a Power Level CC notification event data dict type."""

    args: PowerLevelNotificationArgsDataType  # required


@dataclass
class PowerLevelNotification(BaseNotification):
    """Model for a Zwave Node's Power Level CC notification event."""

    data: PowerLevelNotificationDataType = field(repr=False)
    test_node_id: int = field(init=False)
    status: PowerLevelTestStatus = field(init=False)
    acknowledged_frames: int = field(init=False)

    def __post_init__(self) -> None:
        """Post initialize."""
        super().__post_init__()
        self.test_node_id = self.data["args"]["testNodeId"]
        self.status = PowerLevelTestStatus(self.data["args"]["status"])
        self.acknowledged_frames = self.data["args"]["acknowledgedFrames"]


class MultilevelSwitchNotificationArgsDataType(TypedDict, total=False):
    """Represent args for a Multi Level Switch CC notification event data dict type."""

    eventType: int  # required
    eventTypeLabel: str  # required
    direction: str


class MultilevelSwitchNotificationDataType(BaseNotificationDataType):
    """Represent a Multi Level Switch CC notification event data dict type."""

    args: MultilevelSwitchNotificationArgsDataType  # required


@dataclass
class MultilevelSwitchNotification(BaseNotification):
    """Model for a Zwave Node's Multi Level CC notification event."""

    data: MultilevelSwitchNotificationDataType = field(repr=False)
    event_type: MultilevelSwitchCommand = field(init=False)
    event_type_label: str = field(init=False)
    direction: str | None = field(init=False)

    def __post_init__(self) -> None:
        """Post initialize."""
        super().__post_init__()
        self.event_type = MultilevelSwitchCommand(self.data["args"]["eventType"])
        self.event_type_label = self.data["args"]["eventTypeLabel"]
        self.direction = self.data["args"].get("direction")



"""Constants for utility tests."""

from zwave_js_server.const.command_class.lock import (
    ATTR_CODE_SLOT,
    ATTR_IN_USE,
    ATTR_NAME,
    ATTR_USERCODE,
)

CODE_SLOTS = [
    {
        ATTR_CODE_SLOT: 1,
        ATTR_IN_USE: True,
        ATTR_NAME: "User Code (1)",
        ATTR_USERCODE: "**********",
    },
    {
        ATTR_CODE_SLOT: 2,
        ATTR_IN_USE: True,
        ATTR_NAME: "User Code (2)",
        ATTR_USERCODE: "**********",
    },
    {
        ATTR_CODE_SLOT: 3,
        ATTR_IN_USE: True,
        ATTR_NAME: "User Code (3)",
        ATTR_USERCODE: "**********",
    },
    {
        ATTR_CODE_SLOT: 4,
        ATTR_IN_USE: True,
        ATTR_NAME: "User Code (4)",
        ATTR_USERCODE: "7030\n\r",
    },
    {
        ATTR_CODE_SLOT: 5,
        ATTR_IN_USE: False,
        ATTR_NAME: "User Code (5)",
        ATTR_USERCODE: "",
    },
    {
        ATTR_CODE_SLOT: 6,
        ATTR_IN_USE: False,
        ATTR_NAME: "User Code (6)",
        ATTR_USERCODE: "",
    },
    {
        ATTR_CODE_SLOT: 7,
        ATTR_IN_USE: False,
        ATTR_NAME: "User Code (7)",
        ATTR_USERCODE: "",
    },
    {
        ATTR_CODE_SLOT: 8,
        ATTR_IN_USE: False,
        ATTR_NAME: "User Code (8)",
        ATTR_USERCODE: "",
    },
    {
        ATTR_CODE_SLOT: 9,
        ATTR_IN_USE: False,
        ATTR_NAME: "User Code (9)",
        ATTR_USERCODE: "",
    },
    {
        ATTR_CODE_SLOT: 10,
        ATTR_IN_USE: False,
        ATTR_NAME: "User Code (10)",
        ATTR_USERCODE: "",
    },
    {
        ATTR_CODE_SLOT: 11,
        ATTR_IN_USE: False,
        ATTR_NAME: "User Code (11)",
        ATTR_USERCODE: "",
    },
    {
        ATTR_CODE_SLOT: 12,
        ATTR_IN_USE: False,
        ATTR_NAME: "User Code (12)",
        ATTR_USERCODE: "",
    },
    {
        ATTR_CODE_SLOT: 13,
        ATTR_IN_USE: False,
        ATTR_NAME: "User Code (13)",
        ATTR_USERCODE: "",
    },
    {
        ATTR_CODE_SLOT: 14,
        ATTR_IN_USE: False,
        ATTR_NAME: "User Code (14)",
        ATTR_USERCODE: "",
    },
    {
        ATTR_CODE_SLOT: 15,
        ATTR_IN_USE: False,
        ATTR_NAME: "User Code (15)",
        ATTR_USERCODE: "",
    },
    {
        ATTR_CODE_SLOT: 16,
        ATTR_IN_USE: False,
        ATTR_NAME: "User Code (16)",
        ATTR_USERCODE: "",
    },
    {
        ATTR_CODE_SLOT: 17,
        ATTR_IN_USE: False,
        ATTR_NAME: "User Code (17)",
        ATTR_USERCODE: "",
    },
    {
        ATTR_CODE_SLOT: 18,
        ATTR_IN_USE: False,
        ATTR_NAME: "User Code (18)",
        ATTR_USERCODE: "",
    },
    {
        ATTR_CODE_SLOT: 19,
        ATTR_IN_USE: False,
        ATTR_NAME: "User Code (19)",
        ATTR_USERCODE: "",
    },
    {
        ATTR_CODE_SLOT: 20,
        ATTR_IN_USE: False,
        ATTR_NAME: "User Code (20)",
        ATTR_USERCODE: "",
    },
    {
        ATTR_CODE_SLOT: 21,
        ATTR_IN_USE: False,
        ATTR_NAME: "User Code (21)",
        ATTR_USERCODE: "",
    },
    {
        ATTR_CODE_SLOT: 22,
        ATTR_IN_USE: False,
        ATTR_NAME: "User Code (22)",
        ATTR_USERCODE: "",
    },
    {
        ATTR_CODE_SLOT: 23,
        ATTR_IN_USE: False,
        ATTR_NAME: "User Code (23)",
        ATTR_USERCODE: "",
    },
    {
        ATTR_CODE_SLOT: 24,
        ATTR_IN_USE: False,
        ATTR_NAME: "User Code (24)",
        ATTR_USERCODE: "",
    },
    {
        ATTR_CODE_SLOT: 25,
        ATTR_IN_USE: False,
        ATTR_NAME: "User Code (25)",
        ATTR_USERCODE: "",
    },
    {
        ATTR_CODE_SLOT: 26,
        ATTR_IN_USE: False,
        ATTR_NAME: "User Code (26)",
        ATTR_USERCODE: "",
    },
    {
        ATTR_CODE_SLOT: 27,
        ATTR_IN_USE: False,
        ATTR_NAME: "User Code (27)",
        ATTR_USERCODE: "",
    },
    {
        ATTR_CODE_SLOT: 28,
        ATTR_IN_USE: False,
        ATTR_NAME: "User Code (28)",
        ATTR_USERCODE: "",
    },
    {
        ATTR_CODE_SLOT: 29,
        ATTR_IN_USE: False,
        ATTR_NAME: "User Code (29)",
        ATTR_USERCODE: "",
    },
    {
        ATTR_CODE_SLOT: 30,
        ATTR_IN_USE: None,
        ATTR_NAME: "User Code (30)",
        ATTR_USERCODE: None,
    },
]



"""Constants for the Sound Switch CC."""

from __future__ import annotations

from enum import IntEnum

TONE_ID_PROPERTY = "toneId"
DEFAULT_TONE_ID_PROPERTY = "defaultToneId"
DEFAULT_VOLUME_PROPERTY = "defaultVolume"


class ToneID(IntEnum):
    """Enum with all known Sound Switch CC tone IDs."""

    # https://github.com/zwave-js/node-zwave-js/blob/master/packages/zwave-js/src/lib/commandclass/SoundSwitchCC.ts#L71
    OFF = 0
    DEFAULT = 255



"""Test the firmware update helper."""

from unittest.mock import patch

from zwave_js_server.firmware import driver_firmware_update_otw, update_firmware
from zwave_js_server.model.driver.firmware import (
    DriverFirmwareUpdateData,
    DriverFirmwareUpdateStatus,
)
from zwave_js_server.model.node.firmware import (
    NodeFirmwareUpdateData,
    NodeFirmwareUpdateStatus,
)


async def test_update_firmware_guess_format(url, client_session, multisensor_6):
    """Test update_firmware with guessed format."""
    with (
        patch("zwave_js_server.firmware.Client.connect") as connect_mock,
        patch("zwave_js_server.firmware.Client.initialize") as initialize_mock,
        patch("zwave_js_server.firmware.Client.async_send_command") as cmd_mock,
        patch("zwave_js_server.firmware.Client.disconnect") as disconnect_mock,
    ):
        node = multisensor_6
        cmd_mock.return_value = {
            "result": {"status": 255, "success": True, "reInterview": False}
        }
        result = await update_firmware(
            url, node, [NodeFirmwareUpdateData("test", bytes(10))], client_session
        )
        assert result.status == NodeFirmwareUpdateStatus.OK_RESTART_PENDING
        assert result.success
        assert not result.reinterview

        connect_mock.assert_called_once()
        initialize_mock.assert_called_once()
        cmd_mock.assert_called_once_with(
            {
                "command": "node.update_firmware",
                "nodeId": node.node_id,
                "updates": [
                    {
                        "filename": "test",
                        "file": "AAAAAAAAAAAAAA==",
                    }
                ],
            },
            require_schema=29,
        )
        disconnect_mock.assert_called_once()


async def test_update_firmware_known_format_and_target(
    url, client_session, multisensor_6
):
    """Test update_firmware with known format."""
    with (
        patch("zwave_js_server.firmware.Client.connect") as connect_mock,
        patch("zwave_js_server.firmware.Client.initialize") as initialize_mock,
        patch("zwave_js_server.firmware.Client.async_send_command") as cmd_mock,
        patch("zwave_js_server.firmware.Client.disconnect") as disconnect_mock,
    ):
        node = multisensor_6
        cmd_mock.return_value = {
            "result": {"status": 255, "success": True, "reInterview": False}
        }
        result = await update_firmware(
            url=url,
            node=node,
            updates=[NodeFirmwareUpdateData("test", bytes(10), "test", 1)],
            session=client_session,
        )
        assert result.status == NodeFirmwareUpdateStatus.OK_RESTART_PENDING
        assert result.success
        assert not result.reinterview

        connect_mock.assert_called_once()
        initialize_mock.assert_called_once()
        cmd_mock.assert_called_once_with(
            {
                "command": "node.update_firmware",
                "nodeId": node.node_id,
                "updates": [
                    {
                        "filename": "test",
                        "file": "AAAAAAAAAAAAAA==",
                        "fileFormat": "test",
                        "firmwareTarget": 1,
                    }
                ],
            },
            require_schema=29,
        )
        disconnect_mock.assert_called_once()


async def test_driver_firmware_update_otw_guess_format(url, client_session):
    """Test driver_firmware_update_otw with guessed format."""
    with (
        patch("zwave_js_server.firmware.Client.connect") as connect_mock,
        patch("zwave_js_server.firmware.Client.initialize") as initialize_mock,
        patch("zwave_js_server.firmware.Client.async_send_command") as cmd_mock,
        patch("zwave_js_server.firmware.Client.disconnect") as disconnect_mock,
    ):
        cmd_mock.return_value = {"result": {"status": 255, "success": True}}
        result = await driver_firmware_update_otw(
            url, DriverFirmwareUpdateData("test", bytes(10)), client_session
        )
        assert result.status == DriverFirmwareUpdateStatus.OK
        assert result.success

        connect_mock.assert_called_once()
        initialize_mock.assert_called_once()
        cmd_mock.assert_called_once_with(
            {
                "command": "driver.firmware_update_otw",
                "filename": "test",
                "file": "AAAAAAAAAAAAAA==",
            },
            require_schema=29,
        )
        disconnect_mock.assert_called_once()


async def test_driver_firmware_update_otw_known_format_and_target(url, client_session):
    """Test driver_firmware_update_otw with known format."""
    with (
        patch("zwave_js_server.firmware.Client.connect") as connect_mock,
        patch("zwave_js_server.firmware.Client.initialize") as initialize_mock,
        patch("zwave_js_server.firmware.Client.async_send_command") as cmd_mock,
        patch("zwave_js_server.firmware.Client.disconnect") as disconnect_mock,
    ):
        cmd_mock.return_value = {"result": {"status": 255, "success": True}}
        result = await driver_firmware_update_otw(
            url=url,
            firmware_file=DriverFirmwareUpdateData("test", bytes(10), "test"),
            session=client_session,
        )
        assert result.status == DriverFirmwareUpdateStatus.OK
        assert result.success

        connect_mock.assert_called_once()
        initialize_mock.assert_called_once()
        cmd_mock.assert_called_once_with(
            {
                "command": "driver.firmware_update_otw",
                "filename": "test",
                "file": "AAAAAAAAAAAAAA==",
                "fileFormat": "test",
            },
            require_schema=29,
        )
        disconnect_mock.assert_called_once()
